-- HereBeDragons is a data API for the World of Warcraft mapping system

-- HereBeDragons-2.0 is not supported on WoW 7.x or earlier
if select(4, GetBuildInfo()) < 80000 then
    return
end

local MAJOR, MINOR = "HereBeDragons-2.0", 1
assert(LibStub, MAJOR .. " requires LibStub")

local HereBeDragons, oldversion = LibStub:NewLibrary(MAJOR, MINOR)
if not HereBeDragons then return end

local CBH = LibStub("CallbackHandler-1.0")

HereBeDragons.eventFrame       = HereBeDragons.eventFrame or CreateFrame("Frame")

HereBeDragons.mapData          = HereBeDragons.mapData or {}
HereBeDragons.worldMapData     = HereBeDragons.worldMapData or {}
HereBeDragons.transforms       = HereBeDragons.transforms or {}
HereBeDragons.callbacks        = HereBeDragons.callbacks or CBH:New(HereBeDragons, nil, nil, false)

-- Data Constants
local COSMIC_MAP_ID = 946
local WORLD_MAP_ID = 947

-- Lua upvalues
local PI2 = math.pi * 2
local atan2 = math.atan2
local pairs, ipairs = pairs, ipairs
local type = type
local band = bit.band

-- WoW API upvalues
local UnitPosition = UnitPosition
local C_Map = C_Map

-- data table upvalues
local mapData          = HereBeDragons.mapData -- table { width, height, left, top, .instance, .name, .mapType }
local worldMapData     = HereBeDragons.worldMapData -- table { width, height, left, top }
local transforms       = HereBeDragons.transforms

local currentPlayerUIMapID, currentPlayerUIMapType

-- Override instance ids for phased content
local instanceIDOverrides = {
    -- Draenor
    [1152] = 1116, -- Horde Garrison 1
    [1330] = 1116, -- Horde Garrison 2
    [1153] = 1116, -- Horde Garrison 3
    [1154] = 1116, -- Horde Garrison 4 (unused)
    [1158] = 1116, -- Alliance Garrison 1
    [1331] = 1116, -- Alliance Garrison 2
    [1159] = 1116, -- Alliance Garrison 3
    [1160] = 1116, -- Alliance Garrison 4 (unused)
    [1191] = 1116, -- Ashran PvP Zone
    [1203] = 1116, -- Frostfire Finale Scenario
    [1207] = 1116, -- Talador Finale Scenario
    [1277] = 1116, -- Defense of Karabor Scenario (SMV)
    [1402] = 1116, -- Gorgrond Finale Scenario
    [1464] = 1116, -- Tanaan
    [1465] = 1116, -- Tanaan
    -- Legion
    [1478] = 1220, -- Temple of Elune Scenario (Val'Sharah)
    [1495] = 1220, -- Protection Paladin Artifact Scenario (Stormheim)
    [1498] = 1220, -- Havoc Demon Hunter Artifact Scenario (Suramar)
    [1502] = 1220, -- Dalaran Underbelly
    [1533] = 0,    -- Karazhan Artifact Scenario
    [1612] = 1220, -- Feral Druid Artifact Scenario (Suramar)
    [1626] = 1220, -- Suramar Withered Scenario
    [1662] = 1220, -- Suramar Invasion Scenario
}

-- gather map info, but only if this isn't an upgrade (or the upgrade version forces a re-map)
if not oldversion or oldversion < 1 then
    -- wipe old data, if required, otherwise the upgrade path isn't triggered
    if oldversion then
        wipe(mapData)
        wipe(worldMapData)
        wipe(transforms)
    end

    -- map transform data extracted from UIMapAssignment.db2 (see HereBeDragons-Scripts on GitHub)
    -- format: instanceID, newInstanceID, minY, maxY, minX, maxX, offsetY, offsetX
    local transformData = {
        { 530, 1, -6933.33, 533.33, -16000, -8000, 10133.3, 17600 },
        { 530, 0, 4800, 16000, -10133.3, -2666.67, -2400, 2400 },
        { 732, 0, -20000, 20000, -20000, 20000, -1600, 2800 },
        { 1064, 870, 5391, 8148, 3518, 7655, -2134.2, -2286.6 },
        { 1208, 1116, -2666, -2133, -2133, -1600, 10210, 2410 },
        { 1460, 1220, -1066.7, 2133.3, 0, 3200, -2333.9, 966.7 },
    }

    local function processTransforms()
        for _, transform in pairs(transformData) do
            local instanceID, newInstanceID, minY, maxY, minX, maxX, offsetY, offsetX = unpack(transform)
            table.insert(transforms, { instanceID = instanceID, newInstanceID = newInstanceID, minY = minY, maxY = maxY, minX = minX, maxX = maxX, offsetY = offsetY, offsetX = offsetX })
        end
    end

    local function applyMapTransforms(instanceID, left, right, top, bottom)
        for _, transformData in ipairs(transforms) do
            if transformData.instanceID == instanceID then
                if left <= transformData.maxX and right >= transformData.minX and top <= transformData.maxY and bottom >= transformData.minY then
                    instanceID = transformData.newInstanceID
                    left   = left   + transformData.offsetX
                    right  = right  + transformData.offsetX
                    top    = top    + transformData.offsetY
                    bottom = bottom + transformData.offsetY
                    break
                end
            end
        end
        return instanceID, left, right, top, bottom
    end

    -- map data exported from UIMapAssignment.db2 until API is made available
    -- format: uiMapId -> { mapID, Region0, Region1, Region3, Region4 }
    local db2MapData = { [1]={1,-1716.67,-7250,1808.33,-1962.5},[2]={1,-171.833,-4432.5,5.5003,-4166.5},[3]={1,-286.667,-5177.5005,-203.333,-5052.4995},[4]={1,-286.667,-5177.5005,-203.333,-5052.4995},[5]={1,1400.5,-4899,1580.5,-4629},[6]={1,799.25,-4880.5,971.25,-4622.5},[7]={1,-3802.0801,-3245.8301,-168.75,2204.1699},[8]={1,-2496.6699,229.995,-2263.3301,580.005},[9]={1,-1930,-1559,-1436,-818},[10]={1,-2020.83,-5543.75,1810.42,202.083},[11]={1,-870,-2472.5,-490,-1902.5},[12]={1,-11733.2998,-19733.1992,12799.9004,17066.5996},[13]={0,-15973.3438,-22569.2109,11176.3438,18171.9707},[14]={0,-2460.4199,-4604.1699,-141.667,-1127.08},[15]={0,-7900,-4972.9199,-5854.1699,-1902.08},[16]={0,-6330,-3311.25,-5955,-2748.75},[17]={0,-13025,-4856.25,-10583.2998,-1193.75},[18]={0,825,-1485.42,3837.5,3033.3301},[19]={0,2812.1699,-864.9951,2948.8301,-660.0049},[20]={0,1767.92,1930.51,2200.5801,2579.5},[21]={0,-1133.33,-750,1666.67,3450},[22]={0,500,-3883.3301,3366.6699,416.667},[23]={0,1016.67,-6318.75,3704.1699,-2287.5},[24]={0,2270,-5596.25,2565,-5153.75},[25]={0,-1760.42,-3012.5,1481.25,1850},[26]={0,-1100,-5425,1466.67,-1575},[27]={0,-7206.25,-2760.4199,-3941.6699,2137.5},[28]={0,-6240,-85,-6020,245},[29]={0,-5650,-472.75,-5313,32.75},[30]={0,-5242.5,500,-4862.5,1070},[31]={0,-5736.1699,-1917.0049,-5486.8301,-1542.9951},[32]={0,-7587.5,-2554.1699,-6100,-322.917},[33]={0,-7800,-1473.75,-7325,-761.25},[34]={0,-7650,-1400,-7480,-1145},[35]={0,-7593.3413,-1387.51,-7086.6587,-627.486},[36]={0,-9085.4199,-3616.6699,-6985.4199,-464.583},[37]={0,-10254.2002,-1935.42,-7939.5801,1535.42},[38]={0,-9860,41.75,-9700,281.75},[39]={0,-9880,34.25,-9710,289.25},[40]={0,-9200.5,-769.125,-8985,-445.875},[41]={0,-11395.875,-2437.9199,-10847.0391,-1614.67},[42]={0,-11533.333,-3333.3333,-9866.666,-833.3333},[43]={0,-11441.5996,-2450.7744,-11037.9004,-1845.2256},[44]={0,-11215,-2096.25,-10990,-1758.75},[45]={0,-11225,-2155,-10885,-1645},[46]={0,-11035,-2068.75,-10810,-1731.25},[47]={0,-11516.7002,-1866.67,-9716.6699,833.333},[48]={0,-6327.0801,-4752.0801,-4487.5,-1993.75},[49]={0,-10227.0996,-4047.9199,-8514.5801,-1479.17},[50]={0,-13750,-2356.25,-11016.7002,1743.75},[51]={0,-11208.2998,-4589.5801,-9535.4199,-2081.25},[52]={0,-11733.2998,-483.333,-9400,3016.6699},[53]={0,-10605,1833.75,-10430,2096.25},[54]={0,-9983.333,1293.5,-9798.667,1570.5},[55]={0,-11355,1342.5,-11055,1792.5},[56]={0,-4904.1699,-4525,-2147.9199,-389.583},[57]={1,7931.25,-1639.58,11847.9004,4235.4199},[58]={1,10715,660,11035,1140},[59]={1,10025.2998,970.1997,10211.7002,1249.8003},[60]={1,9728.3301,1454.9951,9881.6699,1685.0049},[61]={1,9678.3301,1354.9951,9931.6699,1735.0049},[62]={1,3912.5,-3447.9199,8222.9199,3016.6699},[63]={1,829.167,-4066.6699,4672.9199,1700},[64]={1,-6900,-4833.3301,-3966.6699,-433.333},[65]={1,-529.167,-1997.92,3404.1699,3902.0801},[66]={1,-2545.8301,-262.5,452.083,4233.3301},[67]={1,-1650,2485.5,-1250,3085.5},[68]={1,-1465.6733,2604.99,-1092.3267,3165.01},[69]={1,-7000,-1508.33,-2366.6699,5441.6699},[70]={1,-5533.3301,-6225,-2033.33,-975},[71]={1,-10579.2002,-7275,-5770.8301,-62.5},[72]={1,-8140,-2940,-7640,-2190},[73]={1,-9490,-4497.5,-8900,-3612.5},[74]={1,-8694.1699,-5271.2549,-7955.8301,-4163.7451},[75]={1,-8907.333,-5013.5,-8036.6665,-3707.5},[76]={1,1704.17,-8887.5,5381.25,-3372.9199},[77]={1,3195.8301,-4264.5801,7237.5,1797.92},[78]={1,-8433.3301,-3166.6699,-5966.6699,533.333},[79]={1,-8185,-1366.25,-7930,-983.75},[80]={1,6952.0801,-3689.5801,8491.6699,-1381.25},[81]={1,-8579.1699,-1075,-5872.9199,2983.3301},[82]={1,-6351.6699,-91.2549,-6183.3301,161.255},[83]={1,4693.75,-7141.6699,8793.75,-991.667},[84]={0,-9154.166,-14.584,-7995.834,1722.917},[85]={1,1327.0833,-5245.729,2486.6667,-3506.354},[86]={1,1690.88,-4526.0601,1932.27,-4163.9702},[87]={0,-5096.8457,-1504.2164,-4569.2412,-713.5914},[88]={1,-1545.8333,-527.0833,-849.9999,516.6666},[89]={1,9437.5,1647.9166,10464.583,3187.5},[90]={0,1237.8412,-86.1824,1877.9453,873.1926},[91]={30,-1739.5833,-2456.25,1085.4166,1781.2499},[92]={489,862.4999,895.8333,1627.0833,2041.6666},[93]={529,337.5,102.0833,1508.3333,1858.3333},[94]={530,7758.3301,-9412.5,11041.7002,-4487.5},[95]={530,6066.6699,-8583.3301,8266.6699,-5283.3301},[96]={530,7510,-7480,7710,-7180},[97]={530,-5508.3301,-14570.7998,-2793.75,-10500},[98]={530,-4890,-11744,-4640,-11369},[99]={530,-3325.8301,-12722.5449,-3009.1699,-12247.5547},[100]={530,-1962.5,375,1481.25,5539.5801},[101]={530,-5821.3594,-4468.0391,5821.3594,12996.0391},[102]={530,-1416.67,4447.9199,1935.42,9475},[103]={530,-4314.3711,-12123.1377,-3609.6833,-11066.3672},[104]={530,-5614.5801,-1275,-1947.92,4225},[105]={530,791.667,3420.8301,4408.3301,8845.8301},[106]={530,-2933.3301,-13337.5,-758.333,-10075},[107]={530,-3641.6699,4770.8301,41.6667,10295.7998},[108]={530,-4600,1683.33,-1000,7083.3301},[109]={530,1739.58,-91.6667,5456.25,5483.3301},[110]={530,9346.9385,-7612.2085,10153.709,-6400.75},[111]={530,-2344.7878,4829.0088,-1473.9545,6135.2588},[112]={566,1404.1666,389.5833,2918.75,2660.4165},[113]={571,-1240.89,-8534.25,10593.4004,9217.1504},[114]={571,1054.17,2806.25,4897.9199,8570.8301},[115]={571,1835.42,-1981.25,5575,3627.0801},[116]={571,2016.67,-6360.4199,5516.6699,-1110.42},[117]={571,-914.583,-7443.75,3116.6699,-1397.92},[118]={571,5245.8301,-827.083,9427.0801,5443.75},[119]={571,4383.3301,2572.9199,7287.5,6929.1699},[120]={571,5456.25,-5270.8301,10197.9004,1841.67},[121]={571,4339.5801,-5593.75,7668.75,-600},[122]={530,11350,-8629.166,13568.749,-5302.083},[123]={571,3733.3301,1354.17,5716.6699,4329.1699},[124]={609,979.1666,-7210.4165,3087.5,-4047.9165},[125]={571,5513.3301,222.495,6066.6699,1052.51},[126]={571,5599.8501,352.646,5975.3398,915.87},[127]={571,4687.5,-1279.17,6502.0801,1443.75},[128]={607,720.8333,-956.2499,1883.3333,787.5},[129]={576,64.0755,-708.069,798.263,393.212},[130]={595,1081.25,327.0833,2297.9165,2152.0833},[131]={595,1891.76,731.06,2641.96,1856.36},[132]={619,200.405,-1205.72,848.684,-233.302},[133]={574,25.6665,-310.406,515.388,424.175},[134]={574,-16.3333,-238.156,304.387,242.925},[135]={574,-75.3335,-510.906,415.721,225.675},[136]={575,186.92,-697.559,552.877,-148.623},[137]={575,157.839,-747.558,661.958,8.622},[138]={602,1157.05,-282.549,1534.54,283.686},[139]={602,959.72,-538.549,1431.88,169.688},[140]={599,762.444,206.324,1375.91,1126.52},[141]={616,611.128,1036.71,897.841,1466.78},[142]={578,222.9167,-262.5,1956.2499,2337.5},[143]={578,877.071,787.253,1220.21,1301.96},[144]={578,927.071,712.253,1370.21,1376.96},[145]={578,927.071,787.253,1270.21,1301.96},[146]={578,990.4028,897.259,1186.8701,1191.96},[147]={603,-1022.9166,-1704.1666,1168.75,1583.3333},[148]={603,1392.71,-445.235,1839.01,224.216},[149]={603,1679.04,-674.74,2564.6799,653.721},[150]={603,1612,-315.75,2219,594.75},[151]={603,2122.53,1684.99,3168.8301,3254.45},[152]={603,1834.77,-310.014,2247.75,309.455},[153]={604,1360.4166,166.6667,2122.9165,1310.4166},[154]={604,1465.52,259.887,2068.8701,1164.92},[155]={615,2841.6665,-29.1667,3616.6665,1133.3333},[156]={624,-634.08,-812.519,298.09,585.736},[157]={601,292.142,-30,794.125,722.974},[158]={601,450.474,400,645.79,692.974},[159]={601,395,462.125,640,829.625},[160]={600,-595.86,-927.01,-182.566,-307.069},[161]={600,-595.86,-1002.01,-182.566,-382.069},[162]={533,2886.6101,-3734.1001,3615.8301,-2640.27},[163]={533,2886.6101,-4234.1001,3615.8301,-3140.27},[164]={533,2336,-3787,3136,-2587},[165]={533,2336.6101,-4287.3501,3136.8301,-3087.02},[166]={533,2311.3401,-4400.0898,3691.22,-2330.28},[167]={533,3379.25,-5522.29,3816.54,-4866.3501},[168]={608,1813.35,665.347,1984.17,921.576},[169]={628,-58.3333,-2125,1708.3333,525},[170]={571,8329.1699,-879.167,10781.2998,2797.9199},[171]={650,629.448,437.424,876.106,807.41},[172]={649,446.361,-41.2552,693.019,328.731},[173]={649,433.024,-211.26,926.354,528.736},[174]={648,-129.167,-131.25,2881.25,4383.3301},[175]={648,514.1667,2815,696.8333,3089},[176]={648,825,716.25,1500,1728.75},[177]={648,1716.6,1622.4999,2068.6001,2150.5},[178]={648,1748.4366,1752.51,1966.7633,2080},[179]={654,-2631.25,293.75,-533.333,3439.5801},[180]={654,-1260,720,-980,1140},[181]={654,-1666.99,2459.77,-1500.01,2710.23},[182]={654,-1666.99,2444.7651,-1480.01,2725.2349},[183]={632,4814.5298,1686.03,5779.9302,3134.1299},[184]={658,233.3333,-693.75,1256.25,839.5833},[185]={668,5126.9902,1469.99,5713.0098,2349.01},[186]={631,-764.84,1384.33,138.807,2739.8},[187]={631,-754.667,1631,-43.3333,2698},[188]={631,-580.313,2116.3301,-449.998,2311.8},[189]={631,4012.4099,1993.5699,4528.2202,2767.28},[190]={631,4002.4099,2216.0601,4768.23,3364.8},[191]={631,4455.75,2586.5701,4704.8799,2960.28},[192]={631,410.292,-2271.55,605.799,-1978.29},[193]={631,414.108,-2648.26,579.396,-2400.3301},[194]={648,-8631.25,777.083,-7731.25,2129.1699},[195]={648,-8438.333,1057.5,-8321.667,1232.5},[196]={648,-8568.333,997.5,-8451.667,1172.5},[197]={648,-8700,1130,-8480,1460},[198]={1,3364.5801,-5175,6195.8301,-929.167},[199]={1,-4737.5,-6056.25,204.167,1356.25},[200]={724,2927.0833,150,3429.1665,902.0833},[201]={0,-5887.5,2268.75,-4018.7498,5070.833},[202]={654,-1899.9999,1043.75,-1306.25,1933.3333},[203]={0,-8352.0801,1808.33,-3720.8301,8754.1699},[204]={0,-7622.9165,4158.333,-4906.25,8233.333},[205]={0,-7989.583,1831.2499,-4756.25,6681.2495},[206]={726,1456.25,-283.3333,2266.6665,931.2499},[207]={646,-604.167,-2047.92,2795.8301,3052.0801},[208]={646,-77.0833,-352.812,160,2.8129},[209]={646,-170,-527.5,350,252.5},[210]={0,-15147.9004,-1837.5,-12516.7002,2108.3301},[213]={389,-452.953,-285.993,39.6232,452.871},[217]={0,-2631.25,293.75,-533.333,3439.5801},[218]={0,-1899.9999,1043.75,-1306.25,1933.3333},[219]={209,1129.1666,241.6667,2052.0833,1624.9999},[220]={109,-718.519,-252.271,-255.166,442.758},[221]={48,-677.147,-381.296,-87.667,502.924},[222]={48,-927.147,-581.296,-337.667,302.924},[223]={48,-877.1483,-281.298,-687.6656,2.926},[224]={0,-15333.2998,-3575,-10964.5996,2977.0801},[225]={34,-31.4695,-188.331,220.633,189.822},[226]={90,-694,-277.772,-180.888,491.896},[227]={90,-714,-77.772,-200.888,691.896},[228]={90,-967.333,-127.772,-387.555,741.896},[229]={90,-937.334,72.9993,-357.554,942.669},[230]={70,-391.473,-248.549,204.306,645.119},[231]={70,-57.7735,52.5486,270.607,545.119},[232]={409,459.861,-1395.5601,1303.0601,-130.76},[233]={309,-12560.416,-2570.8333,-11308.333,-693.75},[234]={429,0,0,0,0},[235]={429,200,-387.5,1050,887.5},[236]={429,-150,125,200,650},[237]={429,-150,231.25,175,718.75},[238]={429,-250,325,250,1075},[239]={429,-281.667,-900,251.667,-99.9992},[240]={429,-200,-862.5,450,112.5},[241]={0,-5670.8301,-7708.3301,-2156.25,-2437.5},[242]={230,248.6397,-884.724,1186.6804,522.337},[243]={230,495.3028,-934.724,1500.0103,572.337},[244]={732,-1904.1666,-4.1667,-560.4166,2010.4166},[245]={732,-847.9166,575,377.0833,2412.5},[246]={540,-91.748,-432.6393,617.417,631.1082},[247]={509,-9908.333,522.9166,-8233.333,3035.4165},[248]={249,-223.833,-371.735,98.2458,111.383},[249]={1,-12158.2998,-3752.0801,-8029.1699,2441.6699},[250]={229,-286.828,-876.252,304.398,10.587},[251]={229,-286.828,-876.252,304.398,10.587},[252]={229,-286.828,-876.252,304.398,10.587},[253]={229,-286.828,-876.252,304.398,10.587},[254]={229,-286.828,-876.252,304.398,10.587},[255]={229,-286.828,-876.252,304.398,10.587},[256]={558,-140.076,-414.9832,354.951,327.5572},[257]={558,-210.076,-602.4832,334.951,215.0573},[258]={556,-295.343,-187.6982,173.654,515.7972},[259]={556,-295.343,-187.6982,173.654,515.7972},[260]={555,-498.218,-656.4432,62.7969,185.0792},[261]={542,-65.8767,-504.53,603.136,498.989},[262]={546,-172.8782,-653.943,423.7352,240.977},[263]={545,-425.9422,-716.742,158.5672,160.022},[264]={545,-425.9422,-716.742,158.5672,160.022},[265]={547,-391.652,-836.124,201.72,53.9341},[266]={553,-256.91,-107.6492,248.025,649.7532},[267]={554,-100.8392,-341.78,349.9862,334.458},[268]={554,-37.8393,-341.78,412.986,334.458},[269]={552,-75.4243,-405.105,384.365,284.579},[270]={552,44,-245.287,408.032,300.761},[271]={552,150.833,-422.605,575.289,214.079},[272]={557,-458.786,-546.3511,90.0708,276.9341},[273]={269,-2225,6562.4995,-1500,7649.9995},[274]={560,1572.9166,-477.0833,3127.0833,1854.1666},[275]={761,735.4166,443.75,1604.1666,1745.8333},[276]={730,337.5,6.25,1370.83,1556.25},[277]={755,-11239.583,-1974.9999,-10591.666,-1004.1666},[279]={43,-410.1463,-375.946,214.1703,560.529},[280]={349,550,-827.5,1200,147.5},[281]={349,-208.333,-1158.75,883.333,478.75},[282]={757,125,1032.5,515,1617.5},[283]={645,6.6667,439.932,686.339,1459.4399},[284]={645,6.6667,239.932,686.339,1259.4399},[285]={669,-567.462,-675.585,-0.9996,174.109},[286]={669,-307.462,-750.584,359,249.109},[287]={469,-7727.0698,-1344.05,-7394.1201,-844.622},[288]={469,-7777.0698,-1379.05,-7344.1201,-729.623},[289]={469,-7757.0698,-1369.05,-7324.1201,-719.623},[290]={469,-7637.0698,-1419.05,-7204.1201,-769.623},[291]={36,-337.509,-796.622,35.3335,-237.358},[292]={36,-267.509,-1016.62,65.3333,-517.357},[293]={670,-880.987,-952.78,-301.622,-83.7326},[294]={671,-899.847,-1184.7705,-180.957,-106.4355},[295]={671,-1289.85,-1034.77,-770.955,-256.427},[296]={671,-1402.85,-1267.27,-707.955,-224.928},[297]={644,-1018.17,-304.781,2.9973,1226.97},[298]={644,-934.835,-1115.29,-86.3316,157.465},[299]={644,-884.8403,-403.294,-132.3277,725.475},[300]={129,2209.8501,570.891,2682.55,1279.9399},[301]={47,1858.6801,1322.47,2349.6399,2058.9199},[302]={189,1616.8586,947.986,2030.1814,1567.97},[303]={189,93.9055,-482.464,307.366,-162.273},[304]={189,1600.64,-562.424,2009.1,50.2726},[305]={189,812.4237,1040.6899,1281.2904,1743.99},[306]={289,104.333,-68.6389,317.698,251.41},[307]={289,44.3326,-128.639,337.699,311.41},[308]={289,34.3212,-113.639,307.707,296.439},[309]={289,-66.3212,-174.122,287.707,356.92},[310]={33,-319.883,2003.7699,-84.9296,2356.2},[311]={33,-303.215,2147.5566,-161.597,2359.9834},[312]={33,-193.216,2103.77,-91.5961,2256.2},[313]={33,-193.215,2103.77,-91.5903,2256.2},[314]={33,-193.215,2103.77,-91.5903,2256.2},[315]={33,-182.546,2080.77,-50.2594,2279.2},[316]={33,-278.216,2023.77,-96.5961,2296.2},[317]={329,3338.97,-3617.6799,3809.45,-2911.96},[318]={329,3498.97,-3967.6802,4169.4502,-2961.9597},[319]={531,-9510.9834,138.076,-7659.2871,2915.6201},[320]={531,-8703.5029,1538.0699,-8051.7964,2515.6299},[321]={531,-8720.1699,1738.0698,-8335.1299,2315.6299},[322]={643,-679.448,308.898,-14,1307.0699},[323]={643,-329.448,308.898,336,1307.0699},[324]={725,797.483,384.121,1675.5699,1701.25},[325]={657,-1457.15,-910.935,-111.332,1107.79},[327]={1,-10733.333,-158.3333,-8033.333,3891.6665},[328]={754,-550,25,450,1525},[329]={534,4479.1665,-4024.9998,6145.833,-1525},[330]={565,-12.5,-50,337.5,475},[331]={544,-115.3334,-170.5,255.3333,385.5},[332]={548,-400,-1362.5005,650.002,212.5025},[333]={568,-277.0833,583.3333,568.75,1852.0833},[334]={550,-100,-787.5,950,787.5},[335]={580,1408.33,300,2012.5,1206.25},[336]={580,1580,377.5,1890,842.5},[337]={859,-12639.583,-2733.3333,-11225,-612.5},[338]={861,908.3333,-256.25,1702.0833,933.3333},[339]={564,427.0833,366.6667,949.9999,1150},[340]={564,-240,23.8705,594.833,1276.1201},[341]={564,380,-176,1030,799},[342]={564,400,-191,1070,814},[343]={564,343.333,134.9995,636.667,575.0005},[344]={564,664.1636,-70,1110.8303,600},[345]={564,450,-67.5,920,637.5},[346]={564,606.6667,137.5,843.3333,492.5},[347]={543,-1492.48,1294.72,-1029.4399,1989.28},[348]={585,-27.7958,-304.226,325.7601,226.108},[349]={585,-27.7959,-304.226,325.7601,226.108},[350]={532,-11189.5996,-2225.0244,-10822.9004,-1674.9756},[351]={532,-11189.0029,-2081.4199,-11017.0967,-1823.5601},[352]={532,-11066.2998,-2132.5747,-10836.2002,-1787.4253},[353]={532,-11119.5996,-2190.0244,-10772.9004,-1669.9756},[354]={532,-10969.2998,-1932.58,-10813.2002,-1698.4301},[355]={532,-11190.5996,-2205.7744,-10802.9004,-1624.2256},[356]={532,-11115.5996,-2066.7744,-10987.9004,-1875.2256},[357]={532,-11105.2002,-2037.6803,-11012.2998,-1898.3297},[358]={532,-11459.5996,-2270.0244,-10952.9004,-1509.9756},[359]={532,-11386.333,-2040.13,-11086.167,-1589.88},[360]={532,-11285.0996,-1825.53,-11104.4004,-1554.48},[361]={532,-11444.5996,-2182.5244,-11047.9004,-1587.4756},[362]={532,-11339.0996,-1963.0244,-10986.4004,-1433.9756},[363]={532,-11143,-2032.63,-10979.5,-1787.38},[364]={532,-11113.6328,-2025.08,-10972.8672,-1813.9301},[365]={532,-11097,-2020.13,-11029.5,-1918.88},[366]={532,-11102,-2155.1299,-10874.5,-1813.8799},[367]={720,-633.3333,-868.7499,425,718.75},[369]={720,265,-770,1225,670},[370]={951,3626.8201,6656.6079,4361.0098,7757.8921},[371]={870,-1002.08,-5531.25,3652.0801,1452.08},[372]={870,2222.3367,-1788,2341.0034,-1610},[373]={870,2160,-1881.25,2435,-1468.75},[374]={870,2000,-2652.5,2220,-2322.5},[375]={870,2251.6667,-2608.75,2463.3333,-2291.25},[376]={870,-1520.83,-1245.83,1095.83,2679.1699},[377]={870,475,1051.2498,726.667,1428.7502},[378]={860,6.25,2308.3301,1785.42,4979.1699},[379]={870,1445.83,-1418.75,5618.75,4839.5801},[380]={870,3388.3191,992.477,3721.6809,1492.52},[381]={870,2456.6599,78.7399,2588.3401,276.2601},[382]={870,3044.5,1652,3174.5,1847},[383]={870,2496.6667,1430,2653.3333,1665},[384]={870,2472,1414.75,2649,1680.25},[385]={870,3525,1480,3835,1945},[386]={870,4105,2458.75,4530,3096.25},[387]={870,4175,2660.25,4400,2997.75},[388]={870,729.167,1335.42,4558.3301,7079.1699},[389]={870,2100,4768.75,2425,5256.25},[390]={870,260.416,-52.084,1947.92,2481.25},[391]={870,1624.5,824.5,1770.5,1043.5},[392]={870,1636.1666,804.5,1808.8334,1063.5},[393]={870,726.333,114.9995,899.667,375.0005},[394]={870,693.75,87.5,903.75,402.5},[395]={870,1480,1642.5,1900,2272.5},[396]={870,1610,1755,1850,2115},[397]={968,1404.1666,389.5833,2918.75,2660.4165},[398]={939,2933.3333,-5902.083,3766.6665,-4650},[399]={940,3297.9165,-1252.0833,5327.083,1791.6666},[400]={940,3425,75,3675,450},[401]={938,2500,-1035.4166,4697.9165,2260.4165},[402]={938,2918.75,231.25,3293.75,793.75},[403]={938,3499.9998,534.8983,4077.0801,1400.5188},[404]={938,4229.165,1181.25,4545.8315,1656.25},[405]={938,2570.8301,-290.1094,3035.4199,406.7754},[406]={938,3970.8301,-646.359,4272.9199,-193.224},[407]={974,-4562.5,5410.416,-3322.916,7268.75},[408]={974,-4235,5941.25,-4140,6083.75},[409]={967,-2628.158,-3940.3013,-565.0928,-833.5928},[410]={967,-1890,-2113.75,-1625,-1716.25},[411]={967,-1910,-3262,-1625,-2834.5},[412]={967,13528.2666,13523.9004,13651.7334,13709.0996},[413]={967,0,-0.25,1,1.25},[414]={967,0,-0.25,1,1.25},[415]={967,-12254.9004,11427.0742,-11516,12535.4258},[416]={999,-4175,-5037.5,-3468.75,-3979.166},[417]={998,1522.917,904.167,2083.333,1743.75},[418]={870,-3235.4199,-1739.58,-110.416,2947.9199},[419]={870,-750,-994.375,-575,-731.875},[420]={870,-885.0004,-1135.47,-549.9996,-632.969},[421]={870,-763.333,-1009.37,-621.667,-796.871},[422]={870,-2152.0801,787.5,1416.67,6139.5801},[423]={727,425.5829,-200.253,1035.9202,715.253},[424]={870,-3664.3799,-6762.4399,6679.1602,8752.8604},[425]={0,-9216.6699,-781.25,-8570.8301,187.5},[426]={0,-8686,-306,-8500,-27},[427]={0,-6606.25,14.584,-5962.5,979.167},[428]={0,-6623.3301,241.7549,-6445.6699,508.2451},[429]={960,751.6636,-2865,1118.3303,-2315},[430]={960,788.5,-2576.5,920.5,-2378.5},[431]={1001,790,302.5,1100,767.5},[432]={1001,1035.5,265,1347.5,733},[433]={870,493.75,-981.25,1689.583,812.5},[434]={870,1140,-413.75,1735,478.75},[435]={1004,1000.66,428.99,1229.34,772.01},[436]={1004,660,300.5,1060,900.5},[437]={962,616.6636,1605,1483.3303,2905},[438]={962,1123.3267,2191.24,1271.6733,2413.76},[439]={961,-836.667,1212.4995,-663.333,1472.5005},[440]={961,-836.667,1212.4995,-663.333,1472.5005},[441]={961,-818.3333,1047.5,-591.6667,1387.5},[442]={961,-776.667,1044.9995,-623.333,1275.0005},[443]={959,3502.083,2300,4210.417,3362.5},[444]={959,3566.6667,2505,3773.3333,2815},[445]={959,3610,2932.5,3750,3142.5},[446]={959,3895,2700,4155,3090},[447]={1005,2077.083,-1343.75,2389.583,-875},[448]={1024,1756.25,-2322.916,2372.917,-1397.916},[449]={1035,1522.917,904.167,2083.333,1743.75},[450]={1048,-3168.75,314.584,-2564.584,1220.833},[451]={1050,-2041.666,2933.333,-252.084,5616.666},[452]={1051,1568.75,-222.916,2216.667,747.917},[453]={994,-4360,-3052.5,-3870,-2317.5},[454]={994,-4485,-2905,-4125,-2365},[455]={994,-4776.6699,-3058.7549,-4278.3301,-2311.2451},[456]={996,-1258.334,-3200,-789.584,-2497.916},[457]={1011,1352.083,4747.916,2029.167,5762.5},[458]={1011,1438.33,5009.9951,1631.67,5300.0049},[459]={1011,1438.33,5009.9951,1631.67,5300.0049},[460]={1,10066.666,41.666,11033.333,1491.667},[461]={1,-900,-4991.666,0,-3641.666},[462]={1,-3754.166,-1533.334,-2577.084,233.334},[463]={1,-1729.166,-6297.916,-525,-4491.666},[464]={1,-824.667,-5738.7505,-576.333,-5366.2495},[465]={0,1543.75,1058.333,2270.833,2147.917},[466]={0,1976.6667,1800,2123.3335,2020},[467]={530,9766.666,-6983.334,10833.333,-5383.334},[468]={530,-4816.666,-14633.332,-3604.166,-12814.582},[469]={0,-5960.416,-643.75,-4727.084,1206.25},[470]={0,-5600,584,-5470,779},[471]={1008,3735.8301,874.9906,4194.1699,1562.5004},[472]={1008,4088.3301,1244.9951,4376.6699,1677.5049},[473]={1008,3780,1315,4280,2065},[474]={1009,-2433.3335,0,-1966.6667,700},[475]={1009,-2730.0015,-9.9944,-1769.9985,1430.01},[476]={1007,134.665,-79.501,301.333,170.501},[477]={1007,43.6661,-130.0002,340.333,315.0002},[478]={1007,-87.5,-73.75,87.5,188.75},[479]={1007,-87.5,-73.75,87.5,188.75},[480]={1148,3726.6699,418.7549,3868.3301,631.2451},[481]={1030,540,2122.5,970,2767.5},[482]={1030,606.333,2182.2498,830,2517.7502},[483]={1000,-4175,-5037.5,-3468.75,-3979.166},[486]={1103,-1333.334,-1433.334,-587.5,-314.584},[487]={1104,-2120.834,85.416,-891.666,1931.25},[488]={1095,493.75,-981.25,1689.583,812.5},[489]={1095,1140,-413.75,1735,478.75},[490]={1112,427.083,366.667,950,1150},[491]={1112,-240,23.8705,594.833,1276.12},[492]={1112,380,-176,1030,799},[493]={1112,400,-191,1070,814},[494]={1112,343.333,134.9995,636.667,575.0005},[495]={1112,664.1636,-70,1110.8303,600},[496]={1112,450,-67.5,920,637.5},[497]={1112,606.6667,137.5,843.3333,492.5},[498]={1102,-2906.25,793.75,-1352.084,3125},[499]={369,-112.5,2311.5,95.5,2623.5},[500]={369,-172.958,2408.5,-49.6247,2593.5},[501]={1106,5513.3281,222.495,6066.6719,1052.51},[502]={1106,5599.8501,352.6407,5975.3398,915.8753},[503]={1043,1958.33,-4857.5049,2101.6699,-4642.4951},[504]={1064,5391.6699,3518.75,8147.9199,7654.1699},[505]={1064,7210,5292.5,7530,5772.5},[506]={1064,7000,4969,7241.3301,5331},[507]={870,5506.25,216.666,6697.9199,2004.17},[508]={1098,5256.6665,5742.5,6113.3335,7027.5},[509]={1098,5213.3301,4624.9951,6246.6699,6175.0049},[510]={1098,5919.1665,4215,6605.8335,5245},[511]={1098,5918.6865,4018.3601,6312.8735,4609.6401},[512]={1098,5441.6665,4215,6128.3335,5245},[513]={1098,5796.1665,4040,6402.8335,4950},[514]={1098,5445,3695,5985,4505},[515]={1098,5321.6665,4361.25,5733.3335,4978.75},[516]={1126,5391.666,3518.75,8147.916,7654.166},[517]={1126,7210,5292.5,7530,5772.5},[518]={1135,-665.8334,100,355.8333,1632.5},[519]={1105,-533.334,-14.584,189.584,1068.75},[520]={1144,806.25,481.25,1310.417,1239.583},[521]={1144,620,781.25,1000,1351.25},[522]={1131,1243,670,1603,1210},[523]={1130,-5460.416,-687.5,-4904.166,145.834},[524]={1099,2095.833,-4516.666,2560.417,-3818.75},[525]={1116,4197.9199,1391.67,8275,7506.25},[526]={1116,6523.3301,5437.4951,7006.6699,6162.5049},[527]={1116,6537.3301,5479.9951,6960.6699,6115.0049},[528]={1116,6710,5645,6950,6005},[529]={1116,6540,5516,6950,6131},[530]={1116,5475,3296.25,5770,3738.75},[531]={1116,6650,4147.5,7200,4972.5},[532]={1116,6650,4147.5,7200,4972.5},[533]={1116,7360,3237.5,7610,3612.5},[534]={1116,2433.3301,-3541.6699,5933.3301,1708.33},[535]={1116,454.166,-129.166,4429.1699,5833.3301},[536]={1116,1615,1962.5,1965,2487.5},[537]={1116,2005,3092.5,2365,3632.5},[538]={1116,1700,2762.5,2050,3287.5},[539]={1116,-1806.25,-4431.25,2660.4199,2268.75},[540]={1116,1030.75,583.75,1261.75,930.25},[541]={1116,1781.58,-1678.75,2010.08,-1336},[542]={1116,-2656.25,-1379.17,1383.33,4681.25},[543]={1116,4139.5801,-3110.4199,9589.5801,5066.6699},[544]={1116,4326.25,1008.13,4498.75,1266.88},[545]={1116,4330,1013.75,4495,1261.25},[546]={1116,6950,1413.75,7165,1736.25},[547]={1116,6950,1375.75,7215,1773.25},[548]={1116,6499.1665,585,6725.8335,925},[549]={1116,6470.8301,524.995,6784.1699,995.005},[550]={1116,1139.583,3633.333,4922.916,9308.333},[551]={1116,2510,3895,2880,4450},[552]={1116,2989.5801,5181.2451,3235.4199,5550.0049},[553]={1116,2322.5,6075,2722.5,6675},[554]={870,-1366.67,-6483.3301,233.334,-4083.3301},[555]={870,-462.5,-5356.25,-187.5,-4943.75},[556]={1136,806.25,481.25,1310.417,1239.583},[557]={1136,1095.8284,199.995,1729.1716,1150.01},[558]={1136,625,716.25,1000,1278.75},[561]={1136,1690.8784,-4526.0601,1932.2716,-4163.9702},[562]={1136,1500,-4837.5,1900,-4237.5},[563]={1136,1575,-5467.5,2165,-4582.5},[564]={1136,1059.1666,-5700,1865.8334,-4490},[565]={1136,1727.08,-5875,2157.0801,-5230},[566]={1136,1200,-5967.5,1790,-5082.5},[567]={1136,960,-5876.25,1275,-5403.75},[568]={1136,650.833,-6010.0029,1204.17,-5179.9971},[569]={1136,985,-5567.5,1215,-5222.5},[570]={1136,975,-5971.6299,1150,-5709.1299},[571]={1161,-1366.666,-6483.334,233.334,-4083.334},[572]={1116,-3964.1875,-10493.4385,11193,12242.3438},[573]={1175,1675,-918.75,2600,468.75},[574]={1176,1580.4166,-339.756,2019.5834,318.994},[575]={1176,1567.5,-1025,2117.5,-200},[576]={1176,1624.1666,-950,1790.8334,-700},[577]={1265,3235.4199,-3462.5,4841.6699,-1054.17},[578]={1265,4359.9985,-2720.3799,4663.1714,-2265.6201},[579]={1116,1806,-31,2056,344},[580]={1116,1806,-31,2056,344},[581]={1116,1806,-31,2056,344},[582]={1116,1635.417,-137.5,2091.667,545.834},[585]={1116,5356,4305,5616,4695},[586]={1116,5356,4305,5616,4695},[587]={1116,5356,4305,5616,4695},[588]={1116,3495.8301,-5795.8301,5577.0801,-2672.9199},[589]={1116,4210,-4555,4850,-3595},[590]={1116,5345.833,4183.333,5814.583,4885.416},[592]={1277,-89.584,-3639.584,1264.583,-1608.334},[593]={1182,1415,2398.75,2150,3501.25},[594]={1207,2600,2166.667,4333.333,4766.666},[595]={1195,6327.5,-1332.5,7082.5,-200},[596]={1205,-13.8336,3001,611.333,3938.75},[597]={1205,-50,3098.75,485,3901.25},[598]={1205,80,3097.25,455,3659.75},[599]={1205,222.913,2939.3696,682.087,3628.1304},[600]={1205,374.168,3263.1895,674.582,3713.8105},[601]={1209,839.652,1434.6166,1367.99,2227.1235},[602]={1209,941.319,1608.3693,1166.3199,1945.8707},[606]={1208,1593.75,1465.63,1806.25,1784.38},[607]={1208,1593.75,1465.63,1806.25,1784.38},[608]={1208,1560,1505,1740,1775},[609]={1208,1551.6666,1740,1748.3334,2035},[610]={1228,3316.667,7062.5,4254.167,8468.75},[611]={1228,3380,7446.25,3545,7693.75},[612]={1228,3336.6667,7375,3603.3333,7775},[613]={1228,3755,8195.3096,4195,8855.3096},[614]={1228,3725,8056.25,4350,8993.75},[615]={1228,3775,8118.75,4350,8981.25},[616]={1358,-286.828,-876.252,304.398,10.587},[617]={1358,-286.828,-876.252,304.398,10.587},[618]={1358,-286.828,-876.252,304.398,10.587},[619]={1220,-5296.6899,-5738.0801,7262.1299,13100.0996},[620]={1279,10.416,935.417,956.25,2354.167},[622]={1116,3450,-4312.5,3908.333,-3625},[623]={1280,1531.25,-172.916,2977.083,1995.833},[624]={1116,5008.333,-4389.584,5541.666,-3589.584},[626]={1220,-1001.67,4370,-798.333,4675},[627]={1220,-1132.08,4050.0801,-583.244,4873.3301},[628]={1220,-961.664,4258.25,-663.33,4705.75},[629]={1220,-920.667,4321.25,-742.333,4588.75},[630]={1220,-2587.5,3241.667,1806.25,9831.25},[631]={1220,0,6097.5,350,6622.5},[632]={1220,-808.333,6312.5,-375,6962.5},[633]={1220,-1490.5,5992.1299,-1180,6457.8799},[634]={1220,1308.33,-768.75,5177.0801,5033.3301},[635]={1220,4617.9199,0,5087.0801,703.75},[636]={1220,2800.6699,3135.25,3060.3301,3524.75},[637]={1220,1964.17,2520,2100.8301,2725},[638]={1220,1890.67,2420.25,2160.3301,2824.75},[639]={1220,3385.8301,2030,3582.5,2325},[640]={1220,2465,1186.25,2623.3301,1423.75},[641]={1220,1291.667,4095.833,4964.583,9606.25},[642]={1220,3300,7200,3500,7500},[643]={1220,1590,6710,1860,7115},[644]={1220,1670.83,6820,1820.83,7045},[645]={1462,-577.084,727.083,237.5,1947.92},[646]={1220,-2570.834,1066.667,-4.166,4916.666},[647]={1220,-1660.83,818.756,-1349.17,1286.25},[648]={1220,-1660.83,818.756,-1349.17,1286.25},[649]={1463,-66.666,-41.666,706.25,1116.67},[650]={1220,2250,618.75,6962.5,7689.583},[651]={1220,3306.6699,4350.6299,3623.3301,4825.6299},[652]={1220,4003.3301,4238.7476,4220,4563.7524},[653]={1220,5045,4742.5,5415,5297.5},[654]={1220,4730,4447.5,5200,5152.5},[655]={1220,4040,4876.25,4305,5273.75},[656]={1220,4125,4927.5,4335,5242.5},[657]={1220,3275,3893.75,3800,4681.25},[658]={1220,2965,3923.75,3400,4576.25},[659]={1220,5715,4282.5,6045,4777.5},[660]={1220,5300,5009.3799,5675,5571.8799},[661]={1448,3708.3301,-977.084,4183.3301,-264.584},[662]={1448,3763.1252,-530.002,4099.7949,-24.9976},[663]={1448,4008.6665,-490,4218.0034,-175.995},[664]={1448,4145.8301,-508.3365,4437.5,-70.8316},[665]={1448,3540,-485,3880,25},[666]={1448,3950,2220,4340,2805},[667]={1448,4010,1963.75,4275,2361.25},[668]={1448,3570.8301,2024.9951,4054.1699,2750.0049},[669]={1448,-3139.5903,-489.193,-2869.9998,-84.8074},[671]={1475,4347.916,2327.083,5133.333,3506.25},[672]={1481,77.084,285.416,2572.917,4029.167},[673]={1481,1010,1462.5,1260,1837.5},[674]={1481,1875,1456.25,2150,1868.75},[675]={1481,1875,1456.25,2150,1868.75},[676]={1500,-3720.8301,-3091.6699,-1204.17,685.417},[677]={1468,3893.7498,-809.375,4575.0005,212.501},[678]={1468,4094.99,-1000.38,4782.3301,30.6304},[679]={1468,4083.3301,-963.7524,4535,-286.2476},[680]={1220,-256.25,1106.25,3445.8301,6658.3301},[681]={1220,1705,3311.25,2040,3813.75},[682]={1220,360,4188.75,775,4811.25},[683]={1220,65,4435,385,4915},[684]={1220,1990,5052.5,2430,5712.5},[685]={1220,1425,4862.5,2225,6062.5},[686]={1220,1860,2780,2220,3320},[687]={1220,1830,3130,2190,3670},[688]={1220,1808.33,3985,2101.6699,4425},[689]={1220,2530,4417.5,2880,4942.5},[690]={1220,2700,5148.75,3035,5651.25},[691]={1220,1785,3643.75,2160,4206.25},[692]={1220,1990,5052.5,2430,5712.5},[693]={1220,1425,4862.5,2225,6062.5},[694]={1511,3052.083,958.333,4039.583,2439.583},[695]={1479,723.4105,6930,1202.4705,7648.5898},[696]={1495,4443.75,-354.166,5193.75,770.833},[697]={1515,1704.17,-8887.5,5381.25,-3372.9199},[698]={1480,-873.5,1973.75,-480.5,2563.25},[699]={1480,-863.5,1988.75,-570.5,2428.25},[700]={1480,4002.4067,2216.0601,4768.2334,3364.8},[701]={1480,410.292,-2271.5503,605.799,-1978.2898},[702]={1512,1171.8766,1110.4399,1482.1234,1575.8101},[703]={1477,2783.3301,2610.4199,3366.6699,3485.4199},[704]={1477,3025,-137.5,3875,1137.5},[705]={1477,2285,-211.25,3300,1311.25},[706]={1492,7095.8301,6989.5801,7508.3301,7608.3301},[707]={1492,2816.6667,600,3083.3333,1000},[708]={1492,2816.6667,600,3083.3333,1000},[709]={1514,6.25,2308.333,1785.417,4979.166},[710]={1493,4083.3301,-963.752,4535,-286.248},[711]={1493,4096.4902,-1000.38,4783.8301,30.6304},[712]={1493,3893.75,-809.375,4575,212.501},[713]={1456,-4072.9199,3539.5801,-2806.25,5439.5801},[714]={1489,-177.084,829.167,718.75,2172.917},[715]={1540,1395.833,1091.667,1902.083,1850},[716]={1528,-1457.15,-910.9365,-111.332,1107.7905},[717]={1107,2772.9199,633.334,3279.1699,1391.67},[718]={1522,2772.9199,633.334,3279.1699,1391.67},[719]={1519,77.084,285.416,2572.9199,4029.1699},[720]={1519,1435,1243.13,1710,1655.63},[721]={1519,1435,1243.13,1710,1655.63},[723]={1494,1231,-440.25,1472,-78.75},[725]={1572,666.666,358.334,1366.67,1408.33},[726]={1469,666.666,358.334,1366.67,1408.33},[728]={1541,-1258.33,-3200,-789.584,-2497.9199},[729]={1503,2275.0901,-583.091,2850,279.273},[731]={1458,2443.75,939.583,3329.1699,2266.6699},[732]={1544,4509.1665,3837.5,4750.8335,4200},[733]={1466,2604.1699,1116.67,3354.1699,2241.6699},[734]={1513,-965,4569.75,-750,4892.25},[735]={1513,-925,4578.5,-735,4863.5},[736]={1583,3626.8201,6656.6079,4361.0098,7757.8921},[737]={1602,-1457.15,-910.9365,-111.332,1107.7905},[738]={1605,-633.333,-868.75,425,718.75},[739]={1220,4258.3301,4841.6699,4779.1699,5625},[740]={1607,1725.1666,1125.5,1999.8334,1537.5},[741]={1607,1725.1666,1125.5,1999.8334,1537.5},[742]={1600,-679.448,308.898,-14,1307.0699},[743]={1600,-329.448,308.898,336,1307.0699},[744]={1579,1392.7097,-445.235,1839.0103,224.216},[745]={1579,2122.5269,1684.99,3168.8335,3254.45},[746]={1579,1679.0397,-674.74,2564.6802,653.721},[747]={1220,3687.5,6906.25,4275,7787.5},[748]={1604,-177.084,829.167,718.75,2172.9199},[749]={1516,2968.75,4262,3665.25,5306.75},[750]={1220,3777.0801,3868.75,4397.9199,4800},[751]={1501,3085.8398,7176.8799,3530,7843.1201},[752]={1501,3000,7168.6899,3215,7491.1899},[753]={1501,3124.3301,7267.4951,3267.6699,7482.5049},[754]={1501,3115.9866,7193.75,3337.6533,7526.25},[755]={1501,3120,7289.5,3240,7469.5},[756]={1501,3130,7316.25,3245,7488.75},[757]={1536,-12556.25,-13506.25,-12037.5,-12729.166},[758]={1553,2166.6699,7841.6699,2762.5,8735.4199},[759]={1621,606.6667,137.5,843.3333,492.5},[760]={1608,4383.333,-2914.584,5279.166,-1570.834},[761]={1571,808.333,2893.75,1470.83,3887.5},[762]={1571,839.3737,2975,1106.0404,3375},[763]={1571,820,2960.5,1106,3389.5},[764]={1530,-225,2562.5,625,3837.5},[765]={1530,240,2985.4685,706.042,3684.5315},[766]={1530,-75,2556.25,650,3643.75},[767]={1530,370,2467.5,880,3232.5},[768]={1530,145,3165,375,3510},[769]={1530,349.79,2880,675.21,3368.1299},[770]={1530,199.75,3003,375.25,3266.25},[771]={1530,179.219,3033.7893,340,3274.9607},[772]={1530,195.4167,3000,374.5833,3268.75},[773]={1630,-1904.17,-4.1667,-560.417,2010.42},[774]={1630,-412,-4837.25,-17,-4244.75},[775]={1624,-4314.3701,-12123.0996,-3609.6799,-11066.4004},[776]={1624,-5508.3301,-14570.7998,-2793.75,-10500},[777]={1520,1658.3267,1024.99,2091.6733,1675.01},[778]={1520,1658.3284,1070,2075.0017,1695.01},[790]={1220,-4385.4199,3070.8301,-2493.75,5908.3301},[791]={1625,751.6636,-2865,1118.3303,-2315},[792]={1625,788.5,-2576.5,920.5,-2378.5},[793]={1646,1291.72,4095.8799,4964.5298,9606.2002},[794]={1529,1610.66,2041.59,1977.36,2591.6299},[795]={1529,1733.96,2134.04,1964.0601,2479.1799},[796]={1529,1680.66,2076.5901,2027.36,2596.6299},[797]={1529,1413.96,2226.48,1714.0601,2676.73},[798]={1632,2968.75,4262,3665.25,5306.75},[799]={1623,222.917,-262.5,1956.25,2337.5},[800]={1623,990.4028,897.259,1186.8701,1191.96},[801]={1623,927.071,787.2517,1270.21,1301.9602},[802]={1623,927.071,712.2517,1370.21,1376.9602},[803]={1623,877.071,787.2517,1220.21,1301.9602},[804]={1618,1000.66,428.99,1229.34,772.01},[805]={1618,660,300.5,1060,900.5},[806]={1648,2918.75,5162.5,3945.8301,6702.0801},[807]={1648,2285,-211.25,3300,1311.25},[809]={1651,-11189.5996,-2225.0244,-10822.9004,-1674.9756},[810]={1651,-11189.0029,-2081.4199,-11017.0967,-1823.5601},[811]={1651,-11066.2998,-2132.5747,-10836.2002,-1787.4253},[812]={1651,-11119.5996,-2190.0244,-10772.9004,-1669.9756},[813]={1651,-10969.2998,-1932.58,-10813.2002,-1698.4301},[814]={1651,-11190.5996,-2205.7744,-10802.9004,-1624.2256},[815]={1651,-11115.5996,-2066.7744,-10987.9004,-1875.2256},[816]={1651,-11105.2002,-2037.6803,-11012.2998,-1898.3297},[817]={1651,-11229.5996,-2110.0244,-10942.9004,-1679.9756},[818]={1651,-4800,-2668.75,-4425,-2106.25},[819]={1651,-4850,-3056.25,-4175,-2043.75},[820]={1651,3918,-2445.75,4225,-1985.25},[821]={1651,3660,-2516.25,4015,-1983.75},[822]={1651,3330,-2659.6899,4065,-1557.1899},[823]={1692,233.333,-693.75,1256.25,839.583},[824]={1645,3343.75,-700,3922.917,170.834},[825]={1687,310.842,193.496,935.158,1129.97},[826]={1702,5050,4743.75,5425,5306.25},[827]={1708,3498.97,-3967.6802,4169.4502,-2961.9597},[828]={1689,389.1278,803.71,675.8412,1233.78},[829]={1714,2285,-211.25,3300,1311.25},[830]={1669,177.084,58.334,2654.1699,3772.9199},[831]={1669,353.333,1302.5,550,1597.5},[832]={1669,353.333,1302.5,550,1597.5},[833]={1669,2160,1648.75,2685,2436.25},[834]={1723,-6606.25,14.584,-5962.5,979.167},[835]={1688,-337.509,-796.622,35.3337,-237.358},[836]={1688,-267.509,-1016.6202,65.3333,-517.3568},[837]={1681,337.5,102.084,1508.333,1858.333},[838]={1715,-8600,-2766.666,-6333.334,633.334},[839]={1580,935,-308.5175,1329.6899,283.5175},[840]={1694,-714,-77.772,-200.888,691.896},[841]={1694,-967.3334,-127.772,-387.5547,741.896},[842]={1694,-937.334,72.999,-357.554,942.6691},[843]={1170,4187.5,-620.834,4437.5,-245.834},[844]={1686,337.5,102.083,1508.33,1858.33},[845]={1677,-745,2175.3101,-302.5,2839.0601},[846]={1677,-725,2238.4399,-365,2778.4399},[847]={1677,-675,2295.8101,-352.5,2779.5601},[848]={1677,-725,2279.28,-410,2751.78},[849]={1677,-650,2405.03,-480,2660.03},[850]={1676,5785,-1401.25,6520,-298.75},[851]={1676,5724.9985,-1351.88,6217.5015,-613.125},[852]={1676,5900,-1506.25,6575,-493.75},[853]={1676,6183.4102,-998.251,6455,-590.866},[854]={1676,6272.1699,-1095.12,6675,-490.875},[855]={1676,6440,-1055.5,6810,-500.5},[856]={1676,4250,-1700.15,4716.8701,-999.845},[857]={1734,-550,25,450,1525},[858]={1666,-2570.8301,1066.67,-4.166,4916.6699},[859]={1740,862.5,895.833,1627.08,2041.67},[860]={1736,2927.083,150,3429.167,902.083},[861]={1732,1131.25,-1210.416,2577.083,958.333},[862]={1642,-4262.5,-2754.166,1379.167,5710.416},[863]={1642,25,-1806.25,3527.083,3445.833},[864]={1642,-204.166,-122.916,4337.5,6689.583},[865]={1707,2760,1188.75,2975,1511.25},[866]={1707,2760,1188.75,2975,1511.25},[867]={1705,1332.5,5893.75,1592.5,6283.75},[868]={1704,2710,7432.5,3100,8017.5},[869]={1706,4315,4375,4515,4675},[870]={1706,4315,4375,4515,4675},[871]={1738,1775,-1954.166,4029.167,1427.083},[872]={1693,-866.666,1045.83,-520.834,1564.58},[873]={1693,-818,1046.75,-591,1387.25},[874]={1693,-836.167,1211.9995,-662.833,1472.0005},[875]={1642,-4580.4702,-4938.5801,4530.5,8727.9004},[876]={1643,-3381.22,-5280.0298,5280.0298,7437.5},[877]={1728,2783.3301,2077.1201,3366.6699,2952.1201},[878]={1523,77.084,285.416,2572.9199,4029.1699},[879]={1523,1435,1243.13,1710,1655.63},[880]={1523,1435,1243.13,1710,1655.63},[881]={1729,389.1278,803.71,675.8412,1233.78},[882]={1669,4450,8287.5,6622.9199,11545.7998},[883]={1669,4557.98,9689.75,4785.02,10030.2998},[884]={1669,4557.98,9689.75,4785.02,10030.2998},[885]={1669,-4056.25,7879.1699,-1789.58,11279.2002},[886]={1669,-2734.042,8479.8906,-2510.853,8810.7412},[887]={1669,-2734.04,8479.8896,-2510.8501,8810.7402},[888]={1733,-9625,1400,-9025,2300},[889]={1746,-42.4243,-630.105,417.365,59.579},[890]={1746,77,-470.287,441.032,75.761},[891]={1750,-5508.3301,-14570.7998,-2793.75,-10500},[892]={1750,-4314.3901,-12123.2998,-3609.6599,-11066.2002},[893]={1750,-4115,-12100,-3915,-11800},[894]={1750,-4115,-12100,-3915,-11800},[895]={1643,-2631.25,-2429.166,2310.417,4985.416},[896]={1643,-2560.4199,-433.334,1264.58,5304.1699},[897]={1756,11968.7002,10818.7998,15708.2998,16427.0996},[898]={1756,14325,14218.7998,14800,14931.2998},[902]={1756,8291.7598,14064.4004,9041.96,15189.7002},[903]={1753,5275,9612.5,6304.1699,11156.2998},[904]={1782,-8254.166,433.334,-7143.75,2100},[906]={1804,1318.75,-697.916,2664.583,1320.833},[907]={1803,958.333,2166.6699,1631.25,3175},[908]={1760,1254.167,-1335.416,3277.083,1697.917},[909]={1712,-3775,9368.75,-2610.4199,11114.5996},[910]={1712,-2956.6699,10640,-2683.3301,11050},[911]={1712,-3933.9199,-1750,-3461.0801,-1040.75},[912]={1712,-4370.8301,-11447.9199,-3891.6699,-10729.1797},[913]={1712,-4089.6001,-11029.2002,-3785.3999,-10572.9004},[914]={1712,-10908.333,7900,-10241.667,8900},[915]={1712,-12950,-3503.1401,-12279.0996,-2496.7896},[916]={1712,-12697.7998,-2970.8501,-12490,-2659.1499},[917]={1712,-12893.333,-2875,-12326.667,-2025},[918]={1712,2730,-4788,3024,-4347},[919]={1712,3050,-9779.1299,3245,-9486.6299},[920]={1712,3050,-9779.1299,3245,-9486.6299},[921]={1779,-4237.5,-5089.5801,-3877.0801,-4547.9199},[922]={1779,-3962.5,-8377.0801,-3502.0801,-7687.5},[923]={1779,583.334,477.084,814.583,822.917},[924]={1779,-1947.92,-1716.67,-1679.17,-1314.58},[925]={1779,-1568.75,510.416,-1227.08,1022.92},[926]={1779,-4250,341.666,-3912.5,847.917},[927]={1779,-10104.2002,-5089.5801,-9743.7803,-4547.9199},[928]={1779,5104.1001,-9977.0801,5564.52,-9287.5},[929]={1779,4316.73,6343.7798,4547.98,6689.6201},[930]={1779,5518.6802,-1716.67,5787.4302,-1314.58},[931]={1779,-1568.75,7977.1201,-1227.08,8489.6201},[932]={1779,-10116.5996,2475.0701,-9779.0996,2981.3201},[933]={1774,3108.333,-868.75,3652.083,-54.166},[934]={1763,-1206.7108,1820,-508.2841,2867.6399},[935]={1763,-950,2106.25,-700,2481.25},[936]={1754,-2104.1699,-1368.75,-1277.08,-129.166},[938]={1780,13916.666,6108.333,15054.166,7814.583},[939]={1780,14158.333,14114.583,14881.25,15197.916},[940]={1860,353.333,1302.4998,550,1597.5002},[941]={1860,353.333,1302.4998,550,1597.5002},[942]={1643,1379.167,-2333.334,5162.5,3343.75},[943]={1876,-1925,-2908.3301,-514.584,-793.75},[971]={1865,985.417,1729.167,2468.75,3954.167},[972]={1622,985.417,1729.17,2468.75,3954.1699},[973]={1840,1580,377.5,1890,842.5},[974]={1771,-118.75,-3097.916,364.584,-2372.916},[975]={1771,55,-2817.5,225,-2562.5},[976]={1771,70,-2801.25,225,-2568.75},[977]={1771,70,-2801.25,225,-2568.75},[978]={1771,70,-2801.25,225,-2568.75},[979]={1771,29.4164,-2870,275.583,-2500.75},[980]={1771,70,-2808.75,235,-2561.25},[981]={1813,-454.166,-1400,156.25,-483.334},[985]={0,-16530,-16530,12270,12270},[986]={1,-11870,-13370,12470,10970},[987]={530,-5867,-1600,6400,10670},[988]={571,-3733,-6933,11730,8533},[989]={870,-4431,-5122,7461,6770},[990]={1116,-4771,-6276,11630,10120},[991]={1642,-4234.8701,-2630.29,4262.79,5840.8999},[992]={1643,-3786.8677,-3444.1343,5351.6636,4566.4419},[993]={1220,-5296.6899,-5738.0801,7262.1299,13100.0996},[994]={1669,56.4033,391.767,4229.25,6669.6099},[997]={1760,1254.167,-1335.416,3277.083,1697.917},[998]={1760,1237.8412,-86.1824,1877.9453,873.1926},[1004]={1762,-1240.3108,2440,-679.6842,3280.9399},[1009]={1642,1180,3302.5,1610,3947.5},[1010]={1594,489.584,-4350,1470.833,-2877.084},[1011]={1642,-4580.4702,-4938.5801,4530.5,8727.9004},[1012]={1904,-9154.4102,-17.7225,-7992.3301,1725.39},[1013]={1904,-9154.4102,-17.7225,-7992.3301,1725.39},[1014]={1643,-3381.22,-5280.0298,5280.0298,7437.5},[1015]={1862,-665,-477.5,-365,-27.5},[1016]={1862,-645.8333,-415,-429.1667,-90},[1017]={1862,-555,-338.75,-450,-181.25},[1018]={1862,-605,-347.5,-485,-167.5},[1021]={1929,-8450.8301,1555,-8184.1699,1955},[1022]={1955,-2212.5,1977.083,-1635.416,2841.667},[1032]={1898,8254.166,-5731.25,9125,-4425},[1036]={1893,929.167,-325,1664.583,777.083},[1040]={1864,3320.8301,-1950.0049,3804.1699,-1224.9951},[1029]={1862,-508.3332,-410,-404.9998,-255},[1033]={1892,8822.916,9097.916,9858.333,10650},[1037]={1883,-1404.166,1281.25,-177.084,3120.833},[1041]={1841,545.834,697.917,1227.083,1718.75},[1030]={0,-1666.99,2459.77,-1500.01,2710.23},[1034]={1882,14502.083,14335.416,15566.666,15931.249},[1038]={1877,3187.5,2789.583,3937.5,3916.667},[1042]={1841,1060,1248.13,1300,1608.13},[1031]={0,-1666.99,2444.77,-1480.01,2725.23},[1035]={1897,708.333,172.916,2283.333,2535.417},[1039]={1864,3500,-1831.25,4225,-743.75},[1043]={1877,3760.417,3185.417,4285.416,3972.917},[1044]={1943,-1925,-2908.3301,-514.584,-793.75},}

    -- gather the data of one map (by uiMapID)
    local function processMap(id, data)
        if not id or mapData[id] then return end

        if db2MapData[id] then
            mapData[id] = db2MapData[id]
            mapData[id].name = data.name
            mapData[id].mapType = data.mapType

            local instance, bottom, right, top, left = mapData[id][1], mapData[id][2], mapData[id][3], mapData[id][4], mapData[id][5]
            instance, left, right, top, bottom = applyMapTransforms(instance, left, right, top, bottom)
            mapData[id][1] = left - right
            mapData[id][2] = top - bottom
            mapData[id][3] = left
            mapData[id][4] = top
            mapData[id][5] = nil
            mapData[id].instance = instance
        else
            mapData[id] = {0, 0, 0, 0, instance = -1, name = data.name, mapType = data.mapType}
        end
    end

    local function processMapChildrenRecursive(id)
        local children = C_Map.GetMapChildrenInfo(id)
        if children and #children > 0 then
            for i = 1, #children do
                local id = children[i].mapID
                if id and not mapData[id] then
                    processMap(id, children[i])
                    processMapChildrenRecursive(id)
                end
            end
        end
    end

    local function fixupZones()
        local cosmic = C_Map.GetMapInfo(COSMIC_MAP_ID)
        mapData[COSMIC_MAP_ID] = {0, 0, 0, 0}
        mapData[COSMIC_MAP_ID].instance = -1
        mapData[COSMIC_MAP_ID].name = cosmic.name
        mapData[COSMIC_MAP_ID].mapType = cosmic.mapType

        -- data for the azeroth world map
        worldMapData[0] = { 62265.29, 41508.24, 53684.04, 19228.04 }
        worldMapData[1] = { 62984, 41980.74, 11291.79, 22778.68 }
        worldMapData[571] = { 57793.83, 39996.64, 29422.39, 11586.76 }
        worldMapData[870] = { 56437.39, 37624.93, 28066.28, 33175.82 }
        worldMapData[1220] = { 82098.05, 54732.04, 54211.17, 25798.86 }
        worldMapData[1642] = { 84076.07, 56079.41, 49045.94, 36813.6 }
        worldMapData[1643] = { 75160.1, 50106.35, 55118.2, 25161.42 }
    end

    local function gatherMapData()
        processTransforms()

        processMapChildrenRecursive(COSMIC_MAP_ID)

        fixupZones()
    end

    gatherMapData()
end

-- Transform a set of coordinates based on the defined map transformations
local function applyCoordinateTransforms(x, y, instanceID)
    for _, transformData in ipairs(transforms) do
        if transformData.instanceID == instanceID then
            if transformData.minX <= x and transformData.maxX >= x and transformData.minY <= y and transformData.maxY >= y then
                instanceID = transformData.newInstanceID
                x = x + transformData.offsetX
                y = y + transformData.offsetY
                break
            end
        end
    end
    if instanceIDOverrides[instanceID] then
        instanceID = instanceIDOverrides[instanceID]
    end
    return x, y, instanceID
end

local StartUpdateTimer
local function UpdateCurrentPosition()
    -- retrieve current zone
    local uiMapID = C_Map.GetBestMapForUnit("player")

    if uiMapID ~= currentPlayerUIMapID then
        -- update upvalues and signal callback
        currentPlayerUIMapID, currentPlayerUIMapType = uiMapID, mapData[uiMapID] and mapData[uiMapID].mapType or 0
        HereBeDragons.callbacks:Fire("PlayerZoneChanged", currentPlayerUIMapID, currentPlayerUIMapType)
    end

    -- start a timer to update in micro dungeons since multi-level micro dungeons do not reliably fire events
    if currentPlayerUIMapType == Enum.UIMapType.Micro then
        StartUpdateTimer()
    end
end

-- upgradeable timer callback, don't want to keep calling the old function if the library is upgraded
HereBeDragons.UpdateCurrentPosition = UpdateCurrentPosition
local function UpdateTimerCallback()
    -- signal that the timer ran
    HereBeDragons.updateTimerActive = nil

    -- run update now
    HereBeDragons.UpdateCurrentPosition()
end

function StartUpdateTimer()
    if not HereBeDragons.updateTimerActive then
        -- prevent running multiple timers
        HereBeDragons.updateTimerActive = true

        -- and queue an update
        C_Timer.After(1, UpdateTimerCallback)
    end
end

local function OnEvent(frame, event, ...)
    UpdateCurrentPosition()
end

HereBeDragons.eventFrame:SetScript("OnEvent", OnEvent)
HereBeDragons.eventFrame:UnregisterAllEvents()
HereBeDragons.eventFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA")
HereBeDragons.eventFrame:RegisterEvent("ZONE_CHANGED")
HereBeDragons.eventFrame:RegisterEvent("ZONE_CHANGED_INDOORS")
HereBeDragons.eventFrame:RegisterEvent("NEW_WMO_CHUNK")
HereBeDragons.eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD")

-- if we're loading after entering the world (ie. on demand), update position now
if IsLoggedIn() then
    UpdateCurrentPosition()
end

--- Return the localized zone name for a given uiMapID
-- @param uiMapID uiMapID of the zone
function HereBeDragons:GetLocalizedMap(uiMapID)
    return mapData[uiMapID] and mapData[uiMapID].name or nil
end

--- Get the size of the zone
-- @param uiMapID uiMapID of the zone
-- @return width, height of the zone, in yards
function HereBeDragons:GetZoneSize(uiMapID)
    local data = mapData[uiMapID]
    if not data then return 0, 0 end

    return data[1], data[2]
end

--- Get a list of all map IDs
-- @return array-style table with all known/valid map IDs
function HereBeDragons:GetAllMapIDs()
    local t = {}
    for id in pairs(mapData) do
        table.insert(t, id)
    end
    return t
end

--- Convert local/point coordinates to world coordinates in yards
-- @param x X position in 0-1 point coordinates
-- @param y Y position in 0-1 point coordinates
-- @param zone uiMapID of the zone
function HereBeDragons:GetWorldCoordinatesFromZone(x, y, zone)
    local data = mapData[zone]
    if not data or data[1] == 0 or data[2] == 0 then return nil, nil, nil end
    if not x or not y then return nil, nil, nil end

    local width, height, left, top = data[1], data[2], data[3], data[4]
    x, y = left - width * x, top - height * y

    return x, y, data.instance
end

--- Convert local/point coordinates to world coordinates in yards. The coordinates have to come from the Azeroth World Map
-- @param x X position in 0-1 point coordinates
-- @param y Y position in 0-1 point coordinates
-- @param instance Instance to use for the world coordinates
function HereBeDragons:GetWorldCoordinatesFromAzerothWorldMap(x, y, instance)
    local data = worldMapData[instance]
    if not data or data[1] == 0 or data[2] == 0 then return nil, nil, nil end
    if not x or not y then return nil, nil, nil end

    local width, height, left, top = data[1], data[2], data[3], data[4]
    x, y = left - width * x, top - height * y

    return x, y, instance
end


--- Convert world coordinates to local/point zone coordinates
-- @param x Global X position
-- @param y Global Y position
-- @param zone uiMapID of the zone
-- @param allowOutOfBounds Allow coordinates to go beyond the current map (ie. outside of the 0-1 range), otherwise nil will be returned
function HereBeDragons:GetZoneCoordinatesFromWorld(x, y, zone, allowOutOfBounds)
    local data = mapData[zone]
    if not data or data[1] == 0 or data[2] == 0 then return nil, nil end
    if not x or not y then return nil, nil end

    local width, height, left, top = data[1], data[2], data[3], data[4]
    x, y = (left - x) / width, (top - y) / height

    -- verify the coordinates fall into the zone
    if not allowOutOfBounds and (x < 0 or x > 1 or y < 0 or y > 1) then return nil, nil end

    return x, y
end

--- Convert world coordinates to local/point zone coordinates on the azeroth world map
-- @param x Global X position
-- @param y Global Y position
-- @param instance Instance to translate coordinates from
-- @param allowOutOfBounds Allow coordinates to go beyond the current map (ie. outside of the 0-1 range), otherwise nil will be returned
function HereBeDragons:GetAzerothWorldMapCoordinatesFromWorld(x, y, instance, allowOutOfBounds)
    local data = worldMapData[instance]
    if not data or data[1] == 0 or data[2] == 0 then return nil, nil end
    if not x or not y then return nil, nil end

    local width, height, left, top = data[1], data[2], data[3], data[4]
    x, y = (left - x) / width, (top - y) / height

    -- verify the coordinates fall into the zone
    if not allowOutOfBounds and (x < 0 or x > 1 or y < 0 or y > 1) then return nil, nil end

    return x, y
end

-- Helper function to handle world map coordinate translation
local function TranslateAzerothWorldMapCoordinates(self, x, y, oZone, dZone, allowOutOfBounds)
    if (oZone ~= WORLD_MAP_ID and not mapData[oZone]) or (dZone ~= WORLD_MAP_ID and not mapData[dZone]) then return nil, nil end
    -- determine the instance we're working with
    local instance = (oZone == WORLD_MAP_ID) and mapData[dZone].instance or mapData[oZone].instance
    if not worldMapData[instance] then return nil, nil end

    local data = worldMapData[instance]
    local width, height, left, top = data[1], data[2], data[3], data[4]

    if oZone == WORLD_MAP_ID then
        x, y = self:GetWorldCoordinatesFromAzerothWorldMap(x, y, instance)
        return self:GetZoneCoordinatesFromWorld(x, y, dZone, allowOutOfBounds)
    else
        x, y = self:GetWorldCoordinatesFromZone(x, y, oZone)
        return self:GetAzerothWorldMapCoordinatesFromWorld(x, y, instance, allowOutOfBounds)
    end
end

--- Translate zone coordinates from one zone to another
-- @param x X position in 0-1 point coordinates, relative to the origin zone
-- @param y Y position in 0-1 point coordinates, relative to the origin zone
-- @param oZone Origin Zone, uiMapID
-- @param dZone Destination Zone, uiMapID
-- @param allowOutOfBounds Allow coordinates to go beyond the current map (ie. outside of the 0-1 range), otherwise nil will be returned
function HereBeDragons:TranslateZoneCoordinates(x, y, oZone, dZone, allowOutOfBounds)
    if oZone == dZone then return x, y end

    if oZone == WORLD_MAP_ID or dZone == WORLD_MAP_ID then
        return TranslateAzerothWorldMapCoordinates(self, x, y, oZone, dZone, allowOutOfBounds)
    end

    local xCoord, yCoord, instance = self:GetWorldCoordinatesFromZone(x, y, oZone)
    if not xCoord then return nil, nil end

    local data = mapData[dZone]
    if not data or data.instance ~= instance then return nil, nil end

    return self:GetZoneCoordinatesFromWorld(xCoord, yCoord, dZone, allowOutOfBounds)
end

--- Return the distance from an origin position to a destination position in the same instance/continent.
-- @param instanceID instance ID
-- @param oX origin X
-- @param oY origin Y
-- @param dX destination X
-- @param dY destination Y
-- @return distance, deltaX, deltaY
function HereBeDragons:GetWorldDistance(instanceID, oX, oY, dX, dY)
    if not oX or not oY or not dX or not dY then return nil, nil, nil end
    local deltaX, deltaY = dX - oX, dY - oY
    return (deltaX * deltaX + deltaY * deltaY)^0.5, deltaX, deltaY
end

--- Return the distance between two points on the same continent
-- @param oZone origin zone uiMapID
-- @param oX origin X, in local zone/point coordinates
-- @param oY origin Y, in local zone/point coordinates
-- @param dZone destination zone uiMapID
-- @param dX destination X, in local zone/point coordinates
-- @param dY destination Y, in local zone/point coordinates
-- @return distance, deltaX, deltaY in yards
function HereBeDragons:GetZoneDistance(oZone, oX, oY, dZone, dX, dY)
    local oX, oY, oInstance = self:GetWorldCoordinatesFromZone(oX, oY, oZone)
    if not oX then return nil, nil, nil end

    -- translate dX, dY to the origin zone
    local dX, dY, dInstance = self:GetWorldCoordinatesFromZone(dX, dY, dZone)
    if not dX then return nil, nil, nil end

    if oInstance ~= dInstance then return nil, nil, nil end

    return self:GetWorldDistance(oInstance, oX, oY, dX, dY)
end

--- Return the angle and distance from an origin position to a destination position in the same instance/continent.
-- @param instanceID instance ID
-- @param oX origin X
-- @param oY origin Y
-- @param dX destination X
-- @param dY destination Y
-- @return angle, distance where angle is in radians and distance in yards
function HereBeDragons:GetWorldVector(instanceID, oX, oY, dX, dY)
    local distance, deltaX, deltaY = self:GetWorldDistance(instanceID, oX, oY, dX, dY)
    if not distance then return nil, nil end

    -- calculate the angle from deltaY and deltaX
    local angle = atan2(-deltaX, deltaY)

    -- normalize the angle
    if angle > 0 then
        angle = PI2 - angle
    else
        angle = -angle
    end

    return angle, distance
end

--- Get the current world position of the specified unit
-- The position is transformed to the current continent, if applicable
-- NOTE: The same restrictions as for the UnitPosition() API apply,
-- which means a very limited set of unit ids will actually work.
-- @param unitId Unit Id
-- @return x, y, instanceID
function HereBeDragons:GetUnitWorldPosition(unitId)
    -- get the current position
    local y, x, z, instanceID = UnitPosition(unitId)
    if not x or not y then return nil, nil, instanceIDOverrides[instanceID] or instanceID end

    -- return transformed coordinates
    return applyCoordinateTransforms(x, y, instanceID)
end

--- Get the current world position of the player
-- The position is transformed to the current continent, if applicable
-- @return x, y, instanceID
function HereBeDragons:GetPlayerWorldPosition()
    -- get the current position
    local y, x, z, instanceID = UnitPosition("player")
    if not x or not y then return nil, nil, instanceIDOverrides[instanceID] or instanceID end

    -- return transformed coordinates
    return applyCoordinateTransforms(x, y, instanceID)
end

--- Get the current zone and level of the player
-- The returned mapFile can represent a micro dungeon, if the player currently is inside one.
-- @return uiMapID, mapType
function HereBeDragons:GetPlayerZone()
    return currentPlayerUIMapID, currentPlayerUIMapType
end

--- Get the current position of the player on a zone level
-- The returned values are local point coordinates, 0-1. The mapFile can represent a micro dungeon.
-- @param allowOutOfBounds Allow coordinates to go beyond the current map (ie. outside of the 0-1 range), otherwise nil will be returned
-- @return x, y, uiMapID, mapType
function HereBeDragons:GetPlayerZonePosition(allowOutOfBounds)
    if not currentPlayerUIMapID then return nil, nil, nil, nil end
    local x, y, instanceID = self:GetPlayerWorldPosition()
    if not x or not y then return nil, nil, nil, nil end

    x, y = self:GetZoneCoordinatesFromWorld(x, y, currentPlayerUIMapID, allowOutOfBounds)
    if x and y then
        return x, y, currentPlayerUIMapID, currentPlayerUIMapType
    end
    return nil, nil, nil, nil
end
